#!/usr/bin/env python

import os.path
import os

from twisted.web import client, error

from twisted.internet import reactor

from codeina.events import Event
from codeina.extern.log import log

from codeina.configure import USR_DOWNLOAD as DEFAULT_DOWNLOAD_DIR
import sha

class HTTPProgressDownloader (client.HTTPDownloader):
    """ A simple HTTP Downloader that notifies of its progress
    in the download process.
    """
    def __errorHandler (self, error):
        self.error_callback (self, error)

    def __init__ (self, url, outfile, postdata=None, headers=None, parent=None):

        client.HTTPDownloader.__init__ (self, url.encode('utf-8'), outfile, postdata=postdata, headers=headers) #, timeout=5)
        self.host = self.host.encode ('utf-8')
        self.path = self.path.encode ('utf-8')

        self.progress_callback = Event ()
        self.completed_callback = Event ()
        self.error_callback = Event ()
        self.message_callback = Event ()

        self.total_length = None
        self.total_downloaded = None

        self.outfile = outfile

        self.deferred.addErrback (self.__errorHandler)
        self.status = ''

    def disable (self):
        self.progress_callback = Event ()
        self.completed_callback = Event ()
        self.error_callback = Event ()
        self.message_callback = Event ()

    def noPage (self, reason):
        """ 
        """
        message = reason.getErrorMessage ()
        if self.status == '304': # We got an error message.
            # self.error_callback (self, message) # We notify whoever needs it.
            client.HTTPDownloader.page (self, '')
        elif self.status != '200':
            # self.error_callback (self, message)
            log.info('codeina', 'Problem downloading page %s from host %s: %s' \
              % (self.path, self.host, message))
            client.HTTPDownloader.noPage (self, reason)


    def gotHeaders (self, headers):
        """ Headers are on their way, meaning we shall soon begin to download.
        """
        if self.status == '200': # The response is OK

            if headers.has_key ('content-length'):
                self.total_length = int (headers['content-length'][0])
            else:
                self.total_length = 0
            self.total_downloaded = 0.0

            # Old headers metadata update
            oldheaders = {}
            eTag = headers.get ('etag', '')
            modified = headers.get ('last-modified', '')
            if modified:
                oldheaders['last-modified'] = modified[0]

            self.progress_callback (self.outfile, 0, self.total_length)

        return client.HTTPDownloader.gotHeaders (self, headers)


    def pagePart (self, data):
        
        if self.status == '200':
            self.total_downloaded += len (data)

            # Launch the callback to notify of progress.
            self.progress_callback (self.outfile, self.total_downloaded, self.total_length)

        return client.HTTPDownloader.pagePart (self, data)

    def pageEnd (self):
        self.completed_callback (self)
        return client.HTTPDownloader.pageEnd (self)


class FluendoPackageDownloader (object):
    """ Class that launches downloads for Codeina.
    All the downloads are simultaneous.
    """

    download_limit = 1

    def __init__ (self):
        self.downloadFinished = Event ()
        self.current_downloads = 0
        self.current_factory = None
        self.queue = []
        self.running = False


    def _downloadAborted (self, factory, reason):
        self.current_downloads -= 1
        if len (self.queue) > 0:
            url, outputfile, postdata, partcallback, endcallback, errorcallback = self.queue.pop ()
            self.downloadFile (url, outputfile, postdata, partcallback, endcallback, errorcallback)

    def _downloadComplete (self, factory):
        self.current_downloads -= 1
        if len (self.queue) > 0: # We have to launch another download.
            url, outputfile, postdata, partcallback, endcallback, errorcallback = self.queue.pop ()
            self.downloadFile (url, outputfile, postdata, partcallback, endcallback, errorcallback)

    def _downloadFile (self, url, outputfile, postdata, partcallback=None, endcallback=None, errorcallback=None):
        """ Launch the download of a file.
        @param url: the url of the file
        @param outputfile: The local path where the file should be saved.
        @param postdata: The POST data to send to the http(s) server.
        @param partcallback: A callback that is called everytime a packet is received.
        @param endcallback: A callback that is called when a file is saved.
        @param errorcallback: A callback to be called when the download encounters an error.
        """
        scheme, host, port, path = client._parse (url)

        factory = HTTPProgressDownloader (url, outputfile, postdata=postdata)

        connector = None
        if scheme == 'https':
            from twisted.internet import ssl
            context_factory = ssl.ClientContextFactory ()
            connector = reactor.connectSSL (host, port, factory, context_factory)
        else:
            connector = reactor.connectTCP (host, port, factory)
        
        if partcallback:
            factory.progress_callback.register (partcallback)
        if endcallback:
            factory.completed_callback.register (endcallback)
        if errorcallback:
            factory.error_callback.register (errorcallback)
        factory.connector = connector

        return factory

    def downloadFile (self, url, outputfile, postdata, partcallback=None, endcallback=None, errorcallback=None, sha=None):
        """
        """
        if self.current_downloads < self.download_limit:
            self.current_downloads += 1
            factory = self._downloadFile (url, outputfile, postdata, partcallback, endcallback, errorcallback)
            factory.completed_callback.register (self._downloadComplete)
            factory.error_callback.register (self._downloadAborted) # Error Handling
            self.current_factory = factory
        else:
            self.queue.append ( (url, outputfile, postdata, partcallback, endcallback, errorcallback) )

    def start (self):
        if not self.running:
            reactor.run ()
        else:
            reactor.wakeUp ()
        self.running = True

    def stop (self):
        reactor.stop ()


class HttpDownload (log.Loggable):
    """
    Base HttpDownload class

    override self.download_object to select a user interface (see
    codeina.ui.itext and codeina.ui.igtk for examples)
    """
    download_object = None
    done = False

    def __init__ (self):
        
        pass

    def run (self, url, directory, fn, sha1sum, message=None, post=None):
        self.debug ("Running download for %s (sha1sum=%s)" % (os.path.join (directory, fn), sha1sum))

        # Check if the directory is existing
        if not os.path.exists (directory):
            os.makedirs (directory)

        # Check if we already downloaded the file previously
        if sha1sum and self.verify (os.path.join (directory, fn), sha1sum):
            self.debug ("Already have %s" % (os.path.join (directory, fn)))
            return None

        # Just to comply with the previous version.
        if message:
            print message

        # The correct download class should be set before trying to download
        # anything.
        if not HttpDownload.download_object:
            raise Exception ("download_object not set for HttpDownload")
        
        HttpDownload.download_object.download (url, os.path.join (directory, fn), post)

    def verify (self, f, sha1sum):
        if not sha1sum:
            self.debug ("No sha1sum or filename, can't check (%s)" % (f))
            return True

        # Not downloaded yet !
        if not os.path.isfile (f): return False

        # Open the file and compute its sha1sum
        fs = open (f, 'rb').read ()
        file_sha1sum = sha.new (fs).hexdigest ()
        if file_sha1sum == sha1sum:
            self.debug ("%s matched the provided sha1sum" % (f))
            return True

        self.debug ("%s didn't match the provided sha1sum (%s != %s)" % (f, file_sha1sum, sha1sum))
        return False

